/** 
 *  Odor Analytics Source File
 *
 *  GNU Copyright (C) 2003  Gaspar Sinai <gsinai@yudit.org>  
 *  Copyright (C) 2002  Gaspar Sinai <gsinai@yudit.org>  
 */

#include "TMatrix.h"
#include <stdlib.h>
#include <math.h>


TMatrix::TMatrix (const SFile& in, unsigned int rows, unsigned int cols) 
  : file(in)
{
  rcount = 0;
  ccount = 0;
  darray = 0;
  if (rows==0 || cols==0)
  {
    error = "tm scan: bad dimension for file: "; 
    error.append (file.getName());
  }
  else if (!file.truncate ((unsigned int)(rows*cols)*sizeof(FType) 
    + 2 * sizeof (int)))
  {
    error = "tm scan: can not truncate file: "; 
    error.append (file.getName());
  }
  else
  {
    SFileImage image = file.getFileImage(true);
    unsigned int* array = (unsigned int*) image.array();
    array[0] = rcount = rows;
    array[1] = ccount = cols;
    darray = (FType*) &array[2];
  }
}

TMatrix::TMatrix (const SFile& in) : file (in)
{
  rcount = 0;
  ccount = 0;
  darray = 0;
  if (file.size() < 0)
  {
    error = "Can not read file: ";
    error.append (file.getName());
  }
  else if (file.size() < (long) (2*sizeof(int)))
  {
    error="File is too short: "; 
    error.append (file.getName());
  }
  else
  {
    image = file.getFileImage (false);
    unsigned int* array = (unsigned int*) image.array();
    rcount = array[0];
    ccount = array[1];
    if (rcount == 0)
    {
      error="Matrix file has zero rows: "; 
      error.append (file.getName());
    }
    else if (ccount == 0)
    {
      error="Matrix file has zero columns: "; 
      error.append (file.getName());
    }
    else if (file.size() != (long) (ccount * rcount * sizeof (FType) 
       + 2 * sizeof (int)))
    {
      error="Matrix file is corrupted: ", 
      error.append (file.getName());
    }
    else
    {
      darray = (FType*) &array[2];
    }
  }
  if (darray == 0)
  {
    ccount = 0;
    rcount = 0;
  }
}

TMatrix::TMatrix (const TMatrix& m) : file (m.file)
{
  image = m.image;
  rcount = m.rcount;
  ccount = m.ccount;
  darray = m.darray;
  error = m.error;
}

TMatrix&
TMatrix::operator=(const TMatrix& m)
{
  if (&m != this)
  {
    file = m.file;
    image = m.image;
    rcount = m.rcount;
    ccount = m.ccount;
    darray = m.darray;
    error = m.error;
  }
  return *this;
}

bool
TMatrix::print (FILE* output)
{
  for (unsigned int i=0; i<rcount; i++)
  {
    char* add = "";
    for (unsigned int j=0; j<ccount; j++)
    {
      double vle = (double) get(i,j);
      if (fprintf (output, "%s%g", add, vle)==0)
      {
         error = "print failed.";
         return false;
      }
      add = " ";
    }
    if (fprintf (output, "\n")==0)
    {
      error = "print failed.";
      return false;
    }
  }
  return true;
}

bool
TMatrix::scan (FILE* in)
{
  double fl;
  unsigned int i=0;
  unsigned int ssize = rcount * ccount;
  for (i=0; fscanf (in, "%lf", &fl)==1; i++)
  {
    if (i>=ssize) continue;
    darray[i] = (FType) fl;
  }
  if (i<ssize)
  {
    ccount = rcount = 0;
    error="not enough input."; 
    return false;
  }
  if (i>ssize)
  {
    ccount = rcount = 0;
    error="too much input."; 
    return false;
  }
  return true;
}

SString
TMatrix::getError() const
{
  SString err = error;
  return SString(err);
}

/**
 * Matrix-Matrix Multiplication
 * @param file output file
 * @param matrix1 multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::pm (const SFile& file, const TMatrix& matrix1) const
{
  if (!isOK() || !matrix1.isOK() || columns() != matrix1.rows())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), matrix1.columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register FType sum; 
  register FType* fpp;
  register FType* fpa0;
  register FType* fpa1;
  register FType* fpb0;
  register FType* fpb1;
  register FType* fpr = ret.darray;

  register unsigned int m, n, k;
  register unsigned int mm = rows();
  register unsigned int nn = columns();
  register unsigned int kk = matrix1.columns();
  
  fpa0 = darray;
  fpp = matrix1.darray;
  for (m=0; m<mm; m++, fpa0+=nn)
  {
    fpb0 = fpp;
    for (k=0; k<kk; k++, fpb0++)
    {
      sum = 0;
      fpa1 = fpa0;
      //fpb1 = &matrix1.darray[k];
      fpb1 = fpb0;
      for (n=0; n<nn; n++)
      {
        sum += (*fpa1) * (*fpb1);
        fpa1++;
        fpb1+=kk;
      }
      *fpr = sum;
      fpr++;
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix add. Each element will be added to scalar
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::add (const SFile& file, FType scalar) const
{
  if (!isOK())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, get (r,c) + scalar);
    }
  }
  return TMatrix(ret);
}
/**
 * Matrix power. Each element will be a power of scalar
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::pow (const SFile& file, FType scalar) const
{
  if (!isOK())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, ::pow (get (r,c), scalar));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix logarithm. Each element will be natural logarithm.
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::log (const SFile& file) const
{
  if (!isOK())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, ::log (get (r,c)));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix exp. Each element will be a natural exp.
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::exp (const SFile& file) const
{
  if (!isOK())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, ::exp (get (r,c)));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix abs. Each element will be a abs.
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::abs (const SFile& file) const
{
  if (!isOK())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, ::fabs (get (r,c)));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix-Matrix divison elements by elements.
 * @param file output file
 * @param matrix1 divisor
 * @return this X matrix1
 */
TMatrix
TMatrix::de (const SFile& file, const TMatrix& matrix1) const
{
  if (!isOK() || !matrix1.isOK() || columns() != matrix1.columns() || rows() != matrix1.rows())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      if (matrix1.get(r,c) == 0.0) 
      {
         TMatrix ret = TMatrix(file, 0, 0);
         ret.error = "Division by zero error.";
         return TMatrix(ret);
      }
      ret.set (r, c, get (r,c) / matrix1.get(r,c));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix-Matrix Multiplication elements by elements
 * @param file output file
 * @param matrix1 multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::pe (const SFile& file, const TMatrix& matrix1) const
{
  if (!isOK() || !matrix1.isOK() || columns() != matrix1.columns() || rows() != matrix1.rows())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, get (r,c) * matrix1.get(r,c));
    }
  }
  return TMatrix(ret);
}
/**
 * Matrix Multiplication by scalar
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::ps (const SFile& file, FType scalar) const
{
  if (!isOK())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, get (r,c) * scalar);
    }
  }
  return TMatrix(ret);
}
/**
 * Matrix-Matrix Addition elements by elements
 * @param file output file
 * @param matrix1 multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::se (const SFile& file, const TMatrix& matrix1) const
{
  if (!isOK() || !matrix1.isOK() || columns() != matrix1.columns() || rows() != matrix1.rows())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, get (r,c) + matrix1.get(r,c));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix-Matrix Substraction elements by elements
 * @param file output file
 * @param matrix1 multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::me (const SFile& file, const TMatrix& matrix1) const
{
  if (!isOK() || !matrix1.isOK() || columns() != matrix1.columns() || rows() != matrix1.rows())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, get (r,c) - matrix1.get(r,c));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix Transpondent 
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::tr (const SFile& file) const
{
  if (!isOK())
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, columns(), rows());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (c, r, get (r,c));
    }
  }
  return TMatrix(ret);
}

/**
 * Matrix Transpondent 
 * @param file output file
 * @param scalar multiplier
 * @return this X matrix1
 */
TMatrix
TMatrix::sort (const SFile& file, const SFile& indexfile, unsigned int ind, bool isrow) const
{
  if (!isOK() || isrow ? (ind>=rows()) : (ind>=columns()))
  {
     TMatrix ret = TMatrix(file, 0, 0);
     ret.error = "Matrix dimension error.";
     return TMatrix(ret);
  }
  TMatrix ret = TMatrix(file, rows(), columns());
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  unsigned int dim = isrow?columns():rows();
  unsigned int* index = new unsigned int[dim]; CHECK_NEW(index);
  FType* keys = new FType[dim]; CHECK_NEW(keys);

  unsigned int i;
  for (i=0; i<dim; i++)
  {
    index[i] = i;
    keys[i] = isrow ? get (ind, i) : get (i, ind);
  }
  /* inclusive */
  if (dim > 1) sort (index, keys, 0, (int)dim-1);
  /* add in a sorted manner */
  register unsigned int rr = rows();
  register unsigned int cc = columns();
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, isrow ? get (r,index[c]): get (index[r], c));
    }
  }
  if (indexfile.getName().size())
  {
    TMatrix imat = TMatrix(indexfile, isrow?1:dim, isrow?dim:1);
    if (!imat.isOK())
    {
       return TMatrix(imat);
    }
    for (r=0; r<dim; r++)
    {
      imat.set (isrow?0:r, isrow?r:0, index[r]);
    }
  }
  return TMatrix(ret);
}

/**
 * Quicksort on key movind keys and index
 * @param ind is the array that shold be moved in sync with key
 * @param key contains the keys. 
 * @param from is the starting index
 * @param to is the ending index (inclusive)
 * @author: Gaspar Sinai <gsinai@yudit.org> 2002-05-03
 */
void
TMatrix::sort (unsigned int* ind, FType* key, int from, int to) const
{
  if (from+1 >= to ) return;

  int i=from;
  int j=to;
  int mid= (to+from)/2;
  /* Make a partition */
  while (true)
  {
    /* We compare everything to the pivot */
    while (i<to &&  key[i] <= key[mid]) i++;
    while (j>from && key[j] > key[mid]) j--;

    /* Left index reached right index. Partition is made - mid.  */
    if (i>=j) break;

    FType tt = key[i]; key[i] = key[j]; key[j] = tt;
    unsigned int ti = ind[i]; ind[i] = ind[j]; ind[j] = ti;
    

    /* check went through pivot condition and move pivot */
    if (i==mid) mid=j; /* move to a higher partition */
    else if (j<=mid) mid=i; /* move to a lower partition */
  }
  /* swap mid and j */
  FType tt = key[mid]; key[mid] = key[j]; key[j] = tt;
  unsigned int ti = ind[mid]; ind[mid] = ind[j]; ind[j] = ti;

  /* Sort low partition. Everything from...mid is for sure less */
  sort (ind, key, from, j-1);
  /* Sort high partition. Everything mid...to is for sure greater 
    or equal than mid */
  sort (ind, key, i, to);
}

/**
 * Zero Matrix
 * @param file output file
 * @param _row is number of rows
 * @param _col is number of columns
 * @return a zero matrix
 */
TMatrix
TMatrix::zero (const SFile& file, unsigned int _row, unsigned int _col)
{
  TMatrix ret = TMatrix(file, _row, _col);
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = _col;
  register unsigned int cc = _row;
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, (FType)0);
    }
  }
  return TMatrix(ret);
}
/**
 * Unit Matrix
 * @param file output file
 * @param _rowcol is number of rows (columns)
 * @return a Unit matrix
 */
TMatrix
TMatrix::unit (const SFile& file, unsigned int _rowcol)
{
  TMatrix ret = TMatrix(file, _rowcol, _rowcol);
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = _rowcol;
  register unsigned int cc = _rowcol;
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, (r==c) ? (FType)1 : (FType)0);
    }
  }
  return TMatrix(ret);
}
/**
 * Random Matrix
 * @param file output file
 * @param _row is number of rows
 * @param _col is number of columns
 * @return a zero matrix
 */
TMatrix
TMatrix::rand100 (const SFile& file, unsigned int _row, unsigned int _col)
{
  TMatrix ret = TMatrix(file, _row, _col);
  if (!ret.isOK())
  {
     return TMatrix(ret);
  }
  register unsigned int rr = _col;
  register unsigned int cc = _row;
  register unsigned int r, c;
  for (r=0; r<rr; r++)
  {
    for (c=0; c<cc; c++)
    {
      ret.set (r, c, (FType)(rand() % 10000)/(FType)100);
    }
  }
  return TMatrix(ret);
}
